// Bigger pass-through including normal and texture coords
// No messing with normal vectors = gouraud possible
// Here also with vertex position in view coords = phong possible

#version 150

layout(triangles) in;
layout(triangle_strip, max_vertices = 90) out; // Normal, solid
//layout(line_strip, max_vertices = 90) out; // Wireframe

//in vec2 texCoord[3];   //[3] because this makes a triangle
in vec3 norm[3];
in vec3 surf[3];
in mat4 mx[3]; // projection
//out vec2 texCoordG;
out vec3 exNormalG;
out vec3 surfG;
uniform mat4 projMatrix;
uniform float fix;
uniform float offs; //  = 0.5 for crude tesselation

vec3	P1, P2, P3, N1, N2, N3; // Just renaming
float w12, w21, w32, w23, w13, w31;
vec3 b300, b030, b003, b210, b120, b021, b012, b102, b201;
vec3 E, V, b111;

vec3 n200, n020, n002;
float v12, v23, v31;
vec3 n110, n011, n101;

//float fix = 1.0;

vec3 lookup(float u, float v)
{
	float w = 1.0 - u - v;
	
	float uPow3 = pow(u, 3);
	float vPow3 = pow(v, 3);
	float wPow3 = pow(w, 3);
	float uPow2 = pow(u, 2);
	float vPow2 = pow(v, 2);
	float wPow2 = pow(w, 2);

	vec3 vx = b300 * wPow3 +
            b030 * uPow3 +
            b003 * vPow3 +
            b210 * 3.0 * wPow2 * u +
            b120 * 3.0 * w * uPow2 +
            b201 * 3.0 * wPow2 * v +
            b021 * 3.0 * uPow2 * v +
            b102 * 3.0 * w * vPow2 +
            b012 * 3.0 * u * vPow2 +
            b111 * 6.0 * w * u * v;
	
	return vx;
}

vec3 lookupn(float u, float v)
{
	float w = 1.0 - u - v;
	
	vec3 n = n200 * w*w +
            n020 * u*u +
            n002 * v*v +
            n110*u*w+
            n011*u*v+
            n101*v*w;
            
	return n;
}

void emit(float u, float v)
{
	float w = 1.0 - u - v;
	
    vec3 s = lookup(u, v);
    gl_Position = vec4(s, 1.0) * projMatrix; //    gl_in[i].gl_Position;
//    gl_Position = vec4(P1*u + P2*v + P3*w, 1.0); //    gl_in[i].gl_Position;
//    texCoordG = texCoord[i];
    exNormalG = lookupn(u, v); // norm[i];
//    exNormalG = norm[0]*u + norm[1]*v + norm[2]*w; // cheat
    surfG = s; // surf[0]*u + surf[1]*v + surf[2]*w; // cheat
    EmitVertex();
}

void main()
{
	vec4 middleOfTriangle = vec4(0.0);
  
//	P1 = vec3(gl_in[0].gl_Position)*1.0/gl_in[0].gl_Position.w;
//	P2 = vec3(gl_in[1].gl_Position)*1.0/gl_in[1].gl_Position.w;
//	P3 = vec3(gl_in[2].gl_Position)*1.0/gl_in[2].gl_Position.w;
//	P1 = vec3(gl_in[0].gl_Position);
//	P2 = vec3(gl_in[1].gl_Position);
//	P3 = vec3(gl_in[2].gl_Position);
	P1 = vec3(surf[0]);
	P2 = vec3(surf[1]);
	P3 = vec3(surf[2]);
	N1 = normalize(fix*norm[0]);
	N2 = normalize(fix*norm[1]);
	N3 = normalize(fix*norm[2]);

	w12 = dot(P2 - P1, N1);
	w21 = dot(P1 - P2, N2);
	w32 = dot(P2 - P3, N3);
	w23 = dot(P3 - P2, N2);
	w13 = dot(P3 - P1, N1);
	w31 = dot(P1 - P3, N3);
  
	b300 = P1;
	b030 = P2;
	b003 = P3;
	b210 = (2*P1 + P2 - w12*N1)/3.0;
	b120 = (2*P2 + P1 - w21*N2)/3.0;
	b021 = (2*P2 + P3 - w23*N2)/3.0;
	b012 = (2*P3 + P2 - w32*N3)/3.0;
	b102 = (2*P3 + P1 - w31*N3)/3.0;
	b201 = (2*P1 + P3 - w13*N1)/3.0;
	E = (b210 + b120 + b021 + b012 + b102 + b201) / 6.0;
	V = (P1 + P2 + P3) / 3.0;
	b111 = E + (E - V)/2.0;
	
	n200 = N1;
	n020 = N2;
	n002 = N3;
	v12 = 2.0 * dot(P2-P1, N1+N2)/dot(P2-P1, P2-P1);
	v23 = 2.0 * dot(P3-P2, N2+N3)/dot(P3-P2, P3-P2);
	v31 = 2.0 * dot(P1-P3, N3+N1)/dot(P1-P3, P1-P3);
	n110 = normalize(N1 + N2 - v12*(P2 - P1));
	n011 = normalize(N2 + N3 - v23*(P3 - P2));
	n101 = normalize(N3 + N1 - v31*(P1 - P3));
  
  
  
  for(int i = 0; i < gl_in.length(); i++)
  {
    middleOfTriangle += gl_in[i].gl_Position;
  }
  middleOfTriangle /= gl_in.length();
	
  for(float u = 0.0; u < 1.0; u += offs)
  for(float v = 0.0; u+v < 1.0; v += offs)
  {
// Varje u, v-par motsvarar up till tv trianglar!
		float u0 = u;
		float u1 = u + offs;
		float v0 = v;
		float v1 = v + offs;
// Gr alltid u0,v0 u0v1 u1v0
// Gr ibland den andra: u0v1 u1v1 u1v0

		emit(u0,v0);
		emit(u0,v1);
		emit(u1,v0);
		EndPrimitive();
		if (u1 + v1 <= 1.01)
		{
			emit(u0,v1);
			emit(u1,v1);
			emit(u1,v0);
			EndPrimitive();
		}
		
//    gl_Position = (gl_in[i].gl_Position + middleOfTriangle) / 2.0;
//    gl_Position = gl_in[i].gl_Position;
//    texCoordG = texCoord[i];
//    exNormalG = norm[i];
//    surfG = surf[i];
//    EmitVertex();
	}
}
